<?php

/* vim: set expandtab shiftwidth=4 tabstop=4 softtabstop=4 foldmethod=marker: */

/**
 * Parser for MARC records
 * 
 * This package is based on the PHP MARC package, originally called "php-marc",
 * that is part of the Emilda Project (http://www.emilda.org). Christoffer
 * Landtman generously agreed to make the "php-marc" code available under the
 * GNU LGPL so it could be used as the basis of this PEAR package.
 * 
 * PHP version 5
 *
 * LICENSE: This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @copyright  2003-2006 Oy Realnode Ab, Dan Scott
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @version    CVS: $Id:$
 * @link       http://pear.php.net/package/File_MARC
 */

/**
 * To-dos:
 * 
 * @todo Ensure every function has an explicit return value
 *
 */

require_once 'PEAR/ErrorStack.php';

// {{{ constants

/**
 * MARC records retrieved from a file
 */
define("MARC_SOURCE_FILE", 1);

/**
 * MARC records retrieved from a binary string 
 */
define("MARC_SOURCE_STRING", 2);

/**
 * Hexadecimal value for Subfield indicator
 */
define("MARC_SUBFIELD_INDICATOR", "\x1F");

/**
 * Hexadecimal value for End of Field
 */
define("MARC_END_OF_FIELD", "\x1E");

/**
 * Hexadecimal value for End of Record
 */
define("MARC_END_OF_RECORD", "\x1D");

/**
 * Length of the Directory
 */
define("MARC_DIRECTORY_ENTRY_LEN", 12);

/**
 * Length of the Leader
 */
define("MARC_LEADER_LEN", 24);

/**
 * Add link to the end of the linked list
 */
define("MARC_LINK_ADD_APPEND", 0);

/**
 * Add link before the target link in the linked list
 */
define("MARC_LINK_ADD_BEFORE", 1);

/**
 * Add link after the target link in the linked list
 */
define("MARC_LINK_ADD_AFTER", 2);

/**
 * Define error codes
 */
define("MARC_ERROR_INVALID_FILE", -1);
define("MARC_ERROR_INVALID_SOURCE", -2);
define("MARC_ERROR_INVALID_TERMINATOR", -3);
define("MARC_ERROR_NO_DIRECTORY", -4);
define("MARC_ERROR_INVALID_DIRECTORY_LENGTH", -5);
define("MARC_ERROR_INVALID_DIRECTORY_TAG", -6);
define("MARC_ERROR_INVALID_DIRECTORY_TAG_LENGTH", -7);
define("MARC_ERROR_INVALID_DIRECTORY_OFFSET", -8);
define("MARC_ERROR_INVALID_DIRECTORY", -9);
define("MARC_ERROR_FIELD_EOF", -10);
define("MARC_ERROR_INVALID_INDICATORS", -11);
define("MARC_ERROR_EMPTY_SUBFIELD", -12);
define("MARC_ERROR_INVALID_INDICATOR_REQUEST", -13);
define("MARC_ERROR_INVALID_MODE", -14);
define("MARC_ERROR_INVALID_FIELD", -15);
define("MARC_ERROR_INVALID_SUBFIELD", -16);

// }}}

// {{{ class MARC
/**
 * The main MARC class enables you to read MARC records from a stream or string.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 * @example marc_yaz.php Pretty print a MARC record retrieved through the PECL yaz extension
 */
class MARC {

    // {{{ properties
    /**
     * Array containing raw records
     * 
     * @var array
     */
    protected $_raw;

    /**
     * Error stack
     *
     * @var PEAR::ErrorStack
     */
    protected $_stack;

    /**
     * Current position in the array of records
     *
     * @var integer
     */
    protected $_pointer;
    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Read in MARC records
     *
     * This function reads in MARC record files or strings that
     * contain one or more MARC records.
     *
     * <code>
     * <?php
     * // Retrieve MARC records from a file
     * $journals = new MARC('journals.mrc', MARC_SOURCE_FILE);
     *
     * // Retrieve MARC records from a string (e.g. Z39 query results)
     * $monographs = new MARC($raw_marc, MARC_SOURCE_STRING);
     * ?>
     * </code>
     *
     * @param string $in Name of the file, or a raw MARC string
     * @param int $source Source of the input, either MARC_SOURCE_FILE or MARC_SOURCE_STRING
     * @return PEAR_Error Returns an error if MARC records cannot be read
     */
    function __construct($in, $source=MARC_SOURCE_FILE)
    {
        $this->_stack = new PEAR_ErrorStack('MARC');

        switch ($source) {

        case MARC_SOURCE_FILE:
            if (!file_exists($in)) {
                $this->_stack->push(MARC_ERROR_INVALID_FILE, 'error', array('in' => $in), "Invalid input file '%in%'");
            }
            $input = file($in);
            $recs = explode(MARC_END_OF_RECORD, join("", $input));
            // Append MARC_END_OF_RECORD as we lost it when splitting
            // Last is not record, as it is empty because every record ends
            // with MARC_END_OF_RECORD.
            for ($i = 0; $i < (count($recs)-1); $i++) {
                $this->_raw[] = $recs[$i].MARC_END_OF_RECORD;
            }
            $this->_pointer = 0;
            break;

        case MARC_SOURCE_STRING:
            $this->_raw[] = $in;
            $this->_pointer = 0;
            break;

        default:
            $this->_stack->push(MARC_ERROR_INVALID_SOURCE, 'error', array(), "Invalid source for MARC records");
        }
    }
    // }}}

    // {{{ nextRaw()
    /**
     * Return the next raw MARC record
     *
     * Returns the next raw MARC record, unless all
     * records already have been read.
     *
     * @return string Either a raw record or false
     */
    function nextRaw()
    {
        /**
         * Exit if we are at the end of the file
         */
        if ($this->_pointer >= count($this->_raw)) {
            return false;
        }

        /**
         * Read next line
         */
        $record = $this->_raw[$this->_pointer++];

        // remove illegal stuff that sometimes occurs between records
        $record = preg_replace('/^[\\x0a\\x0d\\x00]+/', "", $record);

        /**
         * Record validation
         */
        if (strlen($record) < 5) {
            $this->_stack->push(MARC_ERROR_INVALID_LENGTH, 'error', array(), "Couldn't find record length");
        }
        $reclen = substr($record,0,5);
        if (!preg_match("/^\d{5}$/", $reclen)) {
            $this->_stack->push(MARC_ERROR_INVALID_LENGTH, 'error', array("reclen" => $reclen), 'Invalid record length "%reclen%"');
        }
        if ($reclen != strlen($record)) {
            $this->_stack->push(MARC_ERROR_INVALID_LENGTH, 'error', array("reclen" => $reclen, "actual" => strlen($record)), 'Specified record length "%reclen%" does not match actual length "%actual%"');
        }

        return $record;
    }
    // }}}

    // {{{ next()
    /**
     * Return next MARC_Record object
     *
     * Decode the next raw MARC record and return
     * @return MARC_Record A MARC_Record object, or false if there are no more records
     */
    function next()
    {
        if ($_raw = $this->nextRaw()) {
            return $this->_decode($_raw);
        } else {
            return false;
        }
    }
    // }}}

    // {{{ _decode()
    /**
     * Decode a given raw MARC record
     *
     * "Port" of Andy Lesters MARC::File::USMARC->decode() Perl function into PHP.
     *
     * @param string $text Raw MARC record
     * @return MARC_Record Decoded MARC_Record object
     */
    private function _decode($text)
    {
        if (!preg_match("/^(\d{5})/", $text, $matches)) {
            $this->_stack->push(MARC_ERROR_INVALID_LENGTH, 'error', array("reclen" => substr($text, 0, 5)), 'Record length "%reclen%" is not numeric');
        }

        $marc = new MARC_Record();

        // Store record length
        $reclen = $matches[1];

        if ($reclen != strlen($text)) {
            $this->_stack->push(MARC_ERROR_INVALID_LENGTH, 'error', array("reclen" => substr($reclen), "actual" => strlen($text)), 'Invalid record length: Leader says "%reclen%" bytes; actual record length is "%actual%"');
        }

        if (substr($text, -1, 1) != MARC_END_OF_RECORD)
            $this->_stack->push(MARC_ERROR_INVALID_TERMINATOR, 'error', array(), 'Invalid record terminator');

        // Store leader
        $marc->setLeader(substr($text, 0, MARC_LEADER_LEN));

        // bytes 12 - 16 of leader give offset to the body of the record
        $data_start = 0 + substr($text, 12, 5);

        // immediately after the leader comes the directory (no separator)
        $dir = substr($text, MARC_LEADER_LEN, $data_start - MARC_LEADER_LEN - 1);  // -1 to allow for \x1e at end of directory

        // character after the directory must be \x1e
        if (substr($text, $data_start-1, 1) != MARC_END_OF_FIELD) {
            $this->_stack->push(MARC_ERROR_NO_DIRECTORY, 'error', array(), 'No directory found');
        }

        // All directory entries 12 bytes long, so length % 12 must be 0
        if (strlen($dir) % MARC_DIRECTORY_ENTRY_LEN != 0) {
            $this->_stack->push(MARC_ERROR_INVALID_DIRECTORY_LENGTH, 'error', array(), 'Invalid directory length');
        }

        // go through all the fields
        $nfields = strlen($dir) / MARC_DIRECTORY_ENTRY_LEN;
        for ($n=0; $n<$nfields; $n++) {
            // As pack returns to key 1, leave place 0 in list empty
            list(, $tag) = unpack("A3", substr($dir, $n*MARC_DIRECTORY_ENTRY_LEN, MARC_DIRECTORY_ENTRY_LEN));
            list(, $len) = unpack("A3/A4", substr($dir, $n*MARC_DIRECTORY_ENTRY_LEN, MARC_DIRECTORY_ENTRY_LEN));
            list(, $offset) = unpack("A3/A4/A5", substr($dir, $n*MARC_DIRECTORY_ENTRY_LEN, MARC_DIRECTORY_ENTRY_LEN));

            // Check directory validity
            if (!preg_match("/^[0-9A-Za-z]{3}$/", $tag)) {
                $this->_stack->push(MARC_ERROR_INVALID_DIRECTORY_TAG, 'error', array("tag" => $tag), 'Invalid tag "%tag%" in directory');
            }
            if (!preg_match("/^\d{4}$/", $len)) {
                $this->_stack->push(MARC_ERROR_INVALID_DIRECTORY_TAG_LENGTH, 'error', array("tag" => $tag, len => $len), 'Invalid length "%len%"in directory for tag "%tag%"');
            }
            if (!preg_match("/^\d{5}$/", $offset)) {
                $this->_stack->push(MARC_ERROR_INVALID_DIRECTORY_OFFSET, 'error', array("tag" => $tag, "offset" => $offset), 'Invalid offset "%offset%" for tag "%tag%" in directory');
            }
            if ($offset + $len > $reclen) {
                $this->_stack->push(MARC_ERROR_INVALID_DIRECTORY, 'error', array("tag" => $tag), 'Directory entry for tag "%tag%" runs past the end of the record');
            }

            $tagdata = substr($text, $data_start + $offset, $len);

            if (substr($tagdata, -1, 1) == MARC_END_OF_FIELD) {
                # get rid of the end-of-tag character
                $tagdata = substr($tagdata, 0, -1);
                --$len;
            } else {
                $this->_stack->push(MARC_ERROR_FIELD_EOF, 'error', array("tag" => $tag), 'Field for tag "%tag%" does not end with an end of field character');
            }

            if (preg_match("/^\d+$/", $tag) and ($tag < 10)) {
                $marc->addField(new MARC_Control_Field($tag, $tagdata));
            } else {
                $subfields = split(MARC_SUBFIELD_INDICATOR, $tagdata);
                $indicators = array_shift($subfields);

                if (strlen($indicators) > 2 || strlen($indicators) == 0) {
                    $this->_stack->push(MARC_ERROR_INVALID_INDICATORS, 'error', array("tag" => $tag, "indicators" => $indicators), 'Invalid indicators "%indicators%" forced to blanks for tag "%tag%"');
                    list($ind1,$ind2) = array(" ", " ");
                } else {
                    $ind1 = substr($indicators, 0, 1);
                    $ind2 = substr($indicators, 1, 1);
                }

                // Split the subfield data into subfield name and data pairs
                $subfield_data = array();
                foreach ($subfields as $subfield) {
                    if (strlen($subfield) > 0) {
                        $subfield_data[] = new MARC_Subfield(substr($subfield, 0, 1), substr($subfield, 1));
                    } else {
                        $this->_stack->push(MARC_ERROR_EMPTY_SUBFIELD, 'error', array("tag" => $tag), 'Entirely empty subfield found in tag "%tag%"');
                    }
                }

                if (!isset($subfield_data)) {
                    $this->_stack->push(MARC_ERROR_EMPTY_SUBFIELD, 'error', array("tag" => $tag), 'No subfield data found in tag "%tag%"');
                }

                $marc->addField(new MARC_Data_Field($tag, $subfield_data, $ind1, $ind2));
            }
        }

        if ($this->_DEBUG and !empty($this->_stack)) {
            print "<pre>";
            print "MARC DECODE WARNINGS:\n\n";
            foreach($this->_stack as $w) {
                print "-->\t$w\n";
            }
            print "</pre>";
        }

        return $marc;
    }
    // }}}

    // {{{ numRecords()
    /**
     * Get the number of records available in this Record
     * @return int The number of records
     */
    function numRecords()
    {
        return count($this->_raw);
    }
    // }}}
}
// }}}

// {{{ class MARC_Record
/**
 * Represents a single MARC record
 * 
 * A MARC record contains a leader and zero or more fields held within a
 * linked list structure. Fields are represented by {@link MARC_Data_Field} objects.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class MARC_Record {

    // {{{ properties
    /**
     * Contains a linked list of {@link MARC_Data_Field} objects for this Record
     * @var MARC_Linked_List
     */
    protected $_fields;

    /**
     * Record leader
     * @var string
     */
    protected $_leader;

    /**
     * Error stack
     *
     * @var PEAR::ErrorStack
     */
    protected $_stack;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Start function
     *
     * Set all variables to defaults to create new MARC_Record object
     * @return true 
     */
    function __construct()
    {
        $this->_stack = new PEAR_ErrorStack('MARC_Record');
        $this->_fields = new MARC_Linked_List();
        $this->_leader = str_repeat(' ', 24);
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $this->_stack;
    }
    // }}}

    // }}}

    // {{{ getLeader()
    /**
     * Get MARC leader
     *
     * Returns the leader for the MARC record. No validation
     * on the specified leader is performed
     *
     * @return string returns the leader
     */
    function getLeader()
    {
        return $this->_leader;
    }
    // }}}

    // {{{ setLeader()
    /**
     * Set MARC record leader
     *
     * Sets the leader for the MARC record. No validation
     * on the specified leader is performed.
     *
     * @param string $leader Leader
     * @return string returns the leader
     */
    function setLeader($leader)
    {
        $this->_leader = $leader;
        return $this->_leader;
    }
    // }}}

    // {{{ addField()
    /**
     * Adds field to MARC record
     *
     * Adds a MARC_Control_Field or MARC_Data_Field object to the existing list
     * of fields according to the specified mode.
     *
     * @param MARC_Field $newField The field to add
     * @param int $mode specifies how to add the field
     * @param MARC_Field $existingField specifies the reference field for locating the new field
     * @return MARC_Field the new MARC_Field object if the operation succeeded, or
     * a PEAR_Error object if the operation failed
     */
    function addField($newField, $mode=MARC_LINK_ADD_APPEND, $existingField=null)
    {
        $objtype = get_class($newField);
        if ($objtype != "MARC_Data_Field" and $objtype != "MARC_Control_Field") {
            if ($existingField != null and get_class($existingField) != "MARC_Data_Field") {
                $this->_stack->push(MARC_ERROR_INVALID_FIELD, 'error', array("field" => get_class($existingField)), 'Target field must be a MARC_Data_Field or MARC_Control_Field object, but was "%field%"');
            }
            $this->_stack->push(MARC_ERROR_INVALID_FIELD, 'error', array("field" => $objtype), 'New field must be a MARC_Data_Field or MARC_Control_Field object, but was "%field%"');
            return false;
        }

        switch ($mode) {
        /* Append as the last field in the record */
        case MARC_LINK_ADD_APPEND:
            $this->_fields->addLink($newField, MARC_LINK_ADD_APPEND);
            break;

        /* Insert before the specified field in the record */
        case MARC_LINK_ADD_BEFORE:
            $this->_fields->addLink($newField, MARC_LINK_ADD_BEFORE, $existingField);
            break;

        /* Insert after the specified field in the record */
        case MARC_LINK_ADD_AFTER:
            $this->_fields->addLink($newField, MARC_LINK_ADD_AFTER, $existingField);
            break;

        default: 
            $this->_stack->push(MARC_ERROR_INVALID_MODE, 'error', array("mode" => $mode), 'addField() mode "%mode%" was not recognized');
            return false;
        }
        return $newField;
    }
    // }}}

    // {{{ _buildDirectory()
    /**
     * Build Record Directory
     *
     * Generate the directory of the Record according to existing data.
     * @return array Array ($fields, $directory, $total, $baseaddress)
     */
    private function _buildDirectory()
    {
        // Vars
        $fields = array();
        $directory = array();

        $dataend = 0;
        // Not sure we can iterate over a MARC_Linked_List
        foreach ($this->_fields as $field) {
            // No empty fields allowed
            if (!$field->isEmpty()) {
                // Get data in raw format
                $str = $field->toRaw();
                $fields[] = $str;

                // Create directory entry
                $len = strlen($str);
                $direntry = sprintf("%03s%04d%05d", $field->getTag(), $len, $dataend);
                $directory[] = $direntry;
                $dataend += $len;
            }
        }

        /**
         * Rules from MARC::Record::USMARC
         */
        $baseaddress =
                MARC_LEADER_LEN +    // better be 24
                (count($directory) * MARC_DIRECTORY_ENTRY_LEN) +
                                // all the directory entries
                1;              // end-of-field marker


        $total =
                $baseaddress +  // stuff before first field
                $dataend +      // Length of the fields
                1;              // End-of-record marker



        return array($fields, $directory, $total, $baseaddress);
    }
    // }}}

    // {{{ setLeaderLengths()
    /**
     * Set MARC record leader lengths
     *
     * Set the Leader lengths of the record according to defaults specified in
     * {@link http://www.loc.gov/marc/bibliographic/ecbdleaderd.html}
     *
     * @param int $reclen
     * @param int $baseaddr
     * @return true 
     */
    function setLeaderLengths($reclen, $baseaddr)
    {
        $this->_leader = substr_replace($this->_leader, sprintf("%05d", $reclen), 0, 5);
        $this->_leader = substr_replace($this->_leader, sprintf("%05d", $baseaddr), 12, 5);
        $this->_leader = substr_replace($this->_leader, '22', 10, 2);
        $this->_leader = substr_replace($this->_leader, '4500', 20, 4);
        return true;
    }
    // }}}

    // {{{ getFields()
    /**
     * Return an array containing all {@link MARC_Data_Field} or
     * {@link MARC_Control_Field} objects * that match the specified tag name.
     * If the tag name is omitted all fields will be returned.
     *
     * @param string $spec tag name (you can use a PCRE)
     * @return array MARC_Data_Field or MARC_Control_Field objects that match the requested tag name
     */
    function getFields($spec=null)
    {
        if ($spec) {
            $matches = array();
            foreach ($this->_fields as $field) {
                if (preg_match("/$spec/", $field->getTag())) {
                    $matches[] = $field;
                }
            }
            return $matches;
        } else {
            return $this->_fields;
        }
    }
    // }}}

    // {{{ getSubfield()
    /**
     * Retrieve subfields from a {@link MARC_Data_Field} object
     *
     * From the specified Field object, return all subfields with a matching code
     *
     * @param MARC_Data_Field $field field object
     * @param string $code Code of subfield (you can use a PCRE)
     * @return MARC_Subfield|array returns a single {@link MARC_Subfield}
     * object, or an array of {@link MARC_Subfield} objects, or false if
     * no codes match $code
     */
    function getSubfield($field, $code)
    {
        return $field->getSubfield($code);
    }
    // }}}

    // {{{ deleteField()
    /**
     * Deletes the requested field from the record.
     *
     * @param MARC_Data_Field $field The field to be deleted
     * @return true
     */
    function deleteField($field)
    {
        $field = null;
        return $true;
    }
    // }}}

    // {{{ deleteAll()
    /**
     * Delete all occurrences of a field matching a tag name from the record.
     *
     * @param string $tag tag for the fields to be deleted
     * @return int number of fields that were deleted
     */
    function deleteAll($tag)
    {
        foreach ($this->getFields() as $field) {
            $cnt = 0;
            if ($field->getTag() == $tag) {
                $field = null;
                $cnt++;
            }
        }
        return $cnt;
    }
    // }}}

    // {{{ duplicate()
    /**
     * Duplicate a record with all its fields and subfields
     *
     * @return MARC_Record Return the duplicated record
     */
    function duplicate()
    {
        $duplicate = new MARC_Record();
        $duplicate->setLeader($this->_leader);

        foreach ($this->getFields() as $field) {
            $duplicate->addField($field);
        }

        return $duplicate;
    }
    // }}}

    // {{{ output methods
    /**
     * ========== OUTPUT METHODS ==========
     */

    // {{{ getFormattedField()
    /**
     * Formatted representation of MARC_Data_Field
     *
     * Format a Field with a sprintf()-like formatting syntax. The formatting
     * codes are the names of the subfields of the field.
     * @param string $tag Field name
     * @param string $format Format string
     * @return string Return formatted string if Field exists, otherwise False
     */
    function getFormattedField($tag, $format)
    {
        $result = "";
        if ($field = $this->getField($tag)) {
            for ($i=0; $i<strlen($format); $i++) {
                $curr = $format[$i];
                if ($curr != "%") {
                    $result[] = $curr;
                } else {
                    $i++;
                    $curr = $format[$i];
                    if ($curr == "%") {
                        $result[] = $curr;
                    } else {
                        $subfield = $this->getSubfield($tag, $curr);
                        $result[] = $subfield->getData();
                    }
                }
            }
            return implode("", $result);
        } else {
            return false;
        }
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return the Record in raw MARC format.
     *
     * @return string Raw MARC data
     */
    function toRaw()
    {
        list($fields, $directory, $reclen, $baseaddress) = $this->_buildDirectory();
        $this->setLeaderLengths($reclen, $baseaddress);

        /**
         * Glue together all parts
         */
        return $this->_leader.implode("", $directory).MARC_END_OF_FIELD.implode("", $fields).MARC_END_OF_RECORD;
    }
    // }}}

    // {{{ __toString()
    /**
     * Return the MARC record in a pretty printed string
     *
     * Return the MARC record in a pretty printed string.
     *
     * @return string Formatted representation of MARC record
     */
    function __toString()
    {
        // Begin output
        $formatted = "LDR ".$this->_leader."\n";
        foreach ($this->_fields as $field) {
            if (!$field->isEmpty()) {
                $formatted .= $field->__toString()."\n";
            }
        }
        return $formatted;
    }
    // }}}

    // }}}
}
// }}}

// {{{ class MARC_Field
/**
 * The MARC_Field class is expected to be extended to reflect the requirements of control and data fields.
 *
 * A MARC field consists of a tag name.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class MARC_Field extends MARC_Link {

    // {{{ properties
    /**
     * The tag name of the Field
     * @var string
     */
    protected $_tag;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * MARC_Field constructor
     *
     * Create a new {@link MARC_Field} object from passed arguments. We have to
     * define placeholders for the arguments required by child classes.
     *
     * @param string $tag tag
     * @param string $subfields placeholder for subfields or control data
     * @param string $ind1 placeholder for first indicator
     * @param string $ind2 placeholder for second indicator
     * @return string Returns warnings, if any, issued during parse
     */
    function __construct($tag, $subfields=null, $ind1=null, $ind2=null) 
    {
        $this->_stack = new PEAR_ErrorStack('MARC_Field');
        $this->_tag = $tag;

        // Check if valid tag
        if (!preg_match("/^[0-9A-Za-z]{3}$/", $tag)) {
            $this->_stack->push(MARC_ERROR_INVALID_TAG, 'error', array("tag" => $tag), 'Tag "%tag%" is not a valid tag.');
        }

    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the data field
     */
    function __destruct() {
        $this->_stack = null;
        $this->_tag = null;
        parent::__destruct();
    }
    // }}}


    // {{{ getTag()
    /**
     * Returns the tag for this {@link MARC_Field} object
     *
     * @return string returns the tag number of the field
     */
    function getTag()
    {
        return $this->_tag;
    }
    // }}}

    // {{{ setTag()
    /**
     * Sets the tag for this {@link MARC_Field} object
     *
     * @param string $tag new value for the tag
     * @return string returns the tag number of the field
     */
    function setTag($tag)
    {
        $this->_tag = $tag;
        return $this->getTag();
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Is empty
     *
     * Checks if the field is empty.
     *
     * @return bool Returns true if the field is empty, otherwise false
     */
    function isEmpty()
    {
        if ($this->getTag()) {
            return false;
        }
        // It is empty
        return true;
    }
    // }}}

    /**
     * ========== OUTPUT METHODS ==========
     */

    // {{{ __toString()
    /**
     * Return Field formatted
     *
     * Return Field as a formatted string.
     *
     * @return string Formatted output of Field
     */
    function __toString()
    {
        return $this->getTag();
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return field in raw MARC format (stub)
     *
     * Return the field formatted in raw MARC for saving into MARC files. This
     * stub method will be extended by the child classes.
     *
     * @return bool Raw MARC
     */
    function toRaw()
    {
        return false;
    }
    // }}}
}
// }}}

// {{{ class MARC_Data_Field
/**
 * The MARC_Data_Field class represents a single field in a MARC record.
 *
 * A MARC field consists of a tag name, two indicators which may be null,
 * and zero or more subfields represented by {@link MARC_Subfield} objects.
 * Subfields are held within a linked list structure.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class MARC_Data_Field extends MARC_Field {

    // {{{ properties
    /**
     * Value of the first indicator
     * @var string
     */
    protected $_ind1;

    /**
     * Value of the second indicator
     * @var string
     */
    protected $_ind2;

    /**
     * Linked list of subfields
     * @var MARC_Linked_List
     */
    protected $_subfields;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Field init function
     *
     * Create a new {@link MARC_Data_Field} object from passed arguments
     *
     * @param string $tag tag
     * @param array $subfields array of {@link MARC_Subfield} objects
     * @param string $ind1 first indicator
     * @param string $ind2 second indicator
     * @return string Returns warnings, if any, issued during parse
     */
    function __construct($tag, $subfields, $ind1=null, $ind2=null) 
    {
        $this->_subfields = new MARC_Linked_List();

        parent::__construct($tag);

        $this->_ind1 = $this->_validateIndicator($ind1);
        $this->_ind2 = $this->_validateIndicator($ind2);

        // we'll let users add subfields after if they so desire
        if ($subfields) {
            $this->addSubfields($subfields);
        }
    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the data field
     */
    function __destruct() {
        $this->_subfields = null;
        $this->_ind1 = null;
        $this->_ind2 = null;
        parent::__destruct();
    }
    // }}}

    // {{{ Explicit destructor: function delete()
    /**
     * Destroys the data field
     */
    function delete() {
        $this->__destruct();
    }
    // }}}

    // {{{ protected function _validateIndicator()
    /**
     * Validate indicator field
     *
     * Validates the value passed in for an indicator
     *
     * @param string $indicator Value of the indicator to be validated
     * @return string Returns warnings, if any, issued during parse
     */
    private function _validateIndicator($indicator) {
        if (!preg_match("/^[0-9A-Za-z ]$/", $indicator)) {
            if ($indicator != "") {
                $this->_stack->push(MARC_ERROR_INVALID_INDICATOR, 'error', array("tag" => $this->getTag(), "indicator" => $indicator), 'Illegal indicator "%indicator%" in field "%tag%" forced to blank');
            }
            $indicator = " ";
        }
        return $indicator;
    }
    // }}}

    // {{{ addSubfield()
    /**
     * Add subfield to a field in a MARC record
     *
     * Given {@link MARC_Subfield} object will be added to the existing list of fields.
     * @param MARC_Subfield $newSubfield subfield to add
     * @param int $mode where to add the new subfield
     * @param MARC_Subfield $existingSubfield reference location for the new subfield
     * @return MARC_Subfield the new {@link MARC_Subfield} object if the operation succeeded, or
     * a PEAR_Error object if the operation failed
     */
    function addSubfield($newSubfield, $mode=MARC_LINK_ADD_APPEND, $existingSubfield=null)
    {
        if (get_class($newSubfield) != "MARC_Subfield") {
            if ($existingSubfield != null and get_class($existingSubfield) != "MARC_Subfield") {
                $this->_stack->push(MARC_ERROR_INVALID_SUBFIELD, 'error', array('class' => get_class($existingSubfield)), 'Target field must be a MARC_Subfield object, but was "%class%"');
            }
            $this->_stack->push(MARC_ERROR_INVALID_SUBFIELD, 'error', array('class' => get_class($newSubfield)), 'New field must be a MARC_Subfield object, but was "%class%"');
            return false;
        }

        switch ($mode) {
        case MARC_LINK_ADD_APPEND:
            $this->_subfields->addLink($newSubfield, MARC_LINK_ADD_APPEND);
            break;

        case MARC_LINK_ADD_BEFORE:
            $this->_subfields->addLink($newSubfield, MARC_LINK_ADD_BEFORE, $existingSubfield);
            break;

        case MARC_LINK_ADD_AFTER:
            $this->_subfields->addLink($newSubfield, MARC_LINK_ADD_AFTER, $existingSubfield);
            break;

        default:
            $this->_stack->push(MARC_ERROR_INVALID_MODE, 'error', array("mode" => $mode), 'addSubfield() mode "%mode%" was not recognized');
            return false;
        }
        return $newSubfield;
    }
    // }}}

    // {{{ addSubfields()
    /**
     * Adds an array of subfields to a {@link MARC_Data_Field} object
     *
     * Appends subfields to existing subfields in the order in which
     * they appear in the array. For finer grained control of the subfield
     * order, use {@link addSubfield()}.
     *
     * @param array $subfields array of {@link MARC_Subfield} objects
     * @return int returns the number of subfields that were added
     */
    function addSubfields($subfields)
    {
        // Just in case someone passes in a single MARC_Subfield instead of an array
        if (get_class($subfields) == "MARC_Subfield") {
            $this->addSubfield($subfields);
            return 1;
        }

        // Add subfields
        $cnt = 0;
        foreach($subfields as $subfield) {
            $this->addSubfield($subfield);
            $cnt++;
        }

        return $cnt;
    }
    // }}}

    // {{{ deleteSubfield()
    function deleteSubfield($subfield)
    {
        $this->_subfields->deleteLink($subfield);
    }
    // }}}

    // {{{ getIndicator()
    /**
     * Get values of indicators
     *
     * @param int $ind number of the indicator (1 or 2)
     * @return string returns indicator value if it exists, otherwise false
     */
    function getIndicator($ind)
    {
        if ($ind == 1) {
            return $this->_ind1;
        } elseif ($ind == 2) {
            return $this->_ind2;
        } else {
            $this->_stack->push(MARC_ERROR_INVALID_INDICATOR_REQUEST, 'error', array("indicator" => $indicator), 'Request for indicator "%indicator%" failed; 1 and 2 are the only valid indicators');
        }
        return false;
    }
    // }}}

    // {{{ setIndicator()
    /**
     * Set value of an indicator
     *
     * @param int $ind number of the indicator (1 or 2)
     * @param string $value value of the indicator
     * @return string returns indicator value if it exists, otherwise false
     */
    function setIndicator($ind, $value)
    {
        switch ($ind) {

        case 1:
            $this->_ind1 = $this->_validateIndicator($value);
            break;

        case 2:
            $this->_ind2 = $this->_validateIndicator($value);
            break;

        default:
            $this->_stack->push(MARC_ERROR_INVALID_INDICATOR_REQUEST, 'error', array("indicator" => $ind), 'Could not set indicator "%indicator%"; 1 and 2 are the only valid indicators');
            return false;
        }

        return getIndicator($ind);
    }
    // }}}

    // {{{ getSubfield()
    /**
     * Returns the {@link MARC_Subfield} object, or array of objects, for a requested code
     *
     * @param string $code subfield code for which the {@link MARC_Subfield} is retrieved
     * @return MARC_Subfield|array returns a single {@link MARC_Subfield} object, or an
     * array of {@link MARC_Subfield} objects, or false if no codes match $code
     */
    function getSubfield($code)
    {
        $results = array();
        // iterate merrily through the subfields looking for the requested code
        $sf = $this->_subfields->rewind();
        do {
            if (!$sf) {
                // there are no subfields at all; get out of here
                break;
            }
            if ($sf->getCode() == $code) {
                $results[] = $sf;
            }
        } while ($sf = $this->_subfields->next());

        switch (count($results)) {
            case 0:
                return false;
                break;

            case 1:
                return $results[0];
                break;

            default:
                return $results;
                break;
        }
    }
    // }}}

    // {{{ getAllSubfields()
    /**
     * Returns an array containing all subfields in this field
     *
     * @return MARC_Linked_List returns a {@link MARC_Linked_List} linked list of {@link MARC_Subfield} objects
     */
    function getAllSubfields()
    {
        return $this->_subfields;
    }
    // }}}

    // {{{ duplicate()
    /**
     * Duplicate field
     *
     * @return Field Duplicated field object
     */
    function duplicate()
    {
        // Create duplicate
        $duplicate = new MARC_Data_Field($this->_tag, null, $this->_ind1, $this->_ind2);

        // Process subfields and add to duplicate
        $_subfields = array();
        foreach($this->_subfields as $subfield) {
            $duplicate->addSubfield($subfield);
        }

        // Return duplicate
        return $duplicate;
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Is empty
     *
     * Checks if the field is empty. Criteria is that it has
     * at least one subfield with data
     * @return bool Returns true if the field is empty, otherwise false
     */
    function isEmpty()
    {
        foreach($this->_subfields as $subfield) {
            // Check if subfield has data
            if (!$subfield->isEmpty()) {
                return false;
            }
        }
        // It is empty
        return true;
    }
    // }}}

    /**
     * ========== OUTPUT METHODS ==========
     */

    // {{{ __toString()
    /**
     * Return Field formatted
     *
     * Return Field as a formatted string.
     *
     * @return string Formatted output of Field
     */
    function __toString()
    {
        // Variables
        $lines = array();
        // Process tag and indicators
        $pre = sprintf("%3s %1s%1s", $this->_tag, $this->_ind1, $this->_ind2);

        // Process subfields
        $this->_subfields->rewind();
        foreach ($this->_subfields as $subfield) {
            $lines[] = sprintf("%6s _%1s%s", $pre, $subfield->getCode(), $subfield->getData());
            $pre = "";
        }

        return join("\n", $lines);
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return Field in Raw MARC
     *
     * Return the Field formatted in Raw MARC for saving into MARC files
     *
     * @return string Raw MARC
     */
    function toRaw()
    {
        $_subfields = array();
        foreach ($this->_subfields as $subfield) {
            if (!$subfield->isEmpty()) {
                $_subfields[] = $subfield->toRaw();
            }
        }
        return $this->_ind1.$this->_ind2.implode("", $_subfields).MARC_END_OF_FIELD;
    }
    // }}}
}
// }}}

// {{{ class MARC_Control_Field
/**
 * The MARC_Control_Field class represents a single control field in a MARC record.
 *
 * A MARC control field consists of a tag name and control data.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class MARC_Control_Field extends MARC_Field {

    // {{{ Properties
    /**
     * Value of field, if field is a Control field
     * @var string
     */
    protected $_data;
    // }}}

    // {{{ Constructor: function __construct()
    /**
     * Field init function
     *
     * Create a new {@link MARC_Control_Field} object from passed arguments
     *
     * @param string $tag tag
     * @param string $data control field data
     * @param string $ind1 placeholder for class strictness
     * @param string $ind2 placeholder for class strictness
     * @return string Returns warnings, if any, issued during parse
     */
    function __construct($tag, $data, $ind1=null, $ind2=null) 
    {
        $this->_data = $data;
        parent::__construct($tag);

    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the control field
     */
    function __destruct() {
        $this->_data = null;
        parent::__destruct();
    }
    // }}}

    // {{{ Explicit destructor: function delete()
    /**
     * Destroys the control field
     */
    function delete() {
        $this->__destruct();
    }
    // }}}


    // {{{ duplicate()
    /**
     * Duplicate field
     *
     * @return Field Duplicated field object
     */
    function duplicate()
    {
        return new MARC_Control_Field($this->_tag, $this->_data);
    }
    // }}}

    // {{{ getData()
    /**
     * Get control field data
     *
     * @return string returns data in control field
     */
    function getData()
    {
        return $this->_data;
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Is empty
     *
     * Checks if the field contains data
     * @return bool Returns true if the field is empty, otherwise false
     */
    function isEmpty()
    {
        return ($this->_data) ? false : true;
    }
    // }}}

    // {{{ setData()
    /**
     * Set control field data
     *
     * @param string $data data for the control field
     * @return bool returns the new data in the control field 
     */
    function setData($data)
    {
        $this->_data = $data;
        return $this->getData();
    }
    // }}}

    // {{{ __toString()
    /**
     * Return as a formatted string
     *
     * Return the control field as a formatted string for pretty printing
     *
     * @return string Formatted output of control Field
     */
    function __toString()
    {
        return sprintf("%3s     %s", $this->_tag, $this->_data);
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return as raw MARC
     *
     * Return the control field formatted in Raw MARC for saving into MARC files
     *
     * @return string Raw MARC
     */
    function toRaw()
    {
        return $this->_data.MARC_END_OF_FIELD;
    }
    // }}}

}
// }}}

// {{{ class MARC_Subfield
/**
 * The MARC_Subfield class represents a single subfield in a MARC record field.
 *
 * Represents a subfield within a MARC field. This class also implements the possibility of
 * duplicate subfields within a single field, i.e. 650 _z Test1 _z Test2 for example. This class
 * also implements all management functions related to a single subfield.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Christoffer Landtman <landtman@realnode.com>
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class MARC_Subfield extends MARC_Link {
    // {{{ properties
    /**
     * Subfield code, e.g. _a, _b
     * @var string
     */
    protected $_code;

    /**
     * Data contained by the subfield
     * @var string
     */
    protected $_data;

    /**
     * Error stack
     *
     * @var PEAR::ErrorStack
     */
    protected $_stack;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * MARC_Subfield constructor
     *
     * Create a new subfield to represent the code and data
     *
     * @param string $code Subfield code
     * @param string $data Subfield data
     */
    function __construct($code, $data)
    {
        $this->_stack = new PEAR_ErrorStack('MARC_Subfield');
        $this->_code = $code;
        $this->_data = $data;
    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Destroys the subfield
     */
    function __destruct() {
        $this->_stack = null;
        $this->_code = null;
        $this->_data = null;
        parent::__destruct();
    }
    // }}}

    // {{{ Explicit destructor: function delete()
    /**
     * Destroys the subfield
     */
    function delete() {
        $this->__destruct();
    }
    // }}}

    // {{{ getCode()
    /**
     * Return code of the subfield
     *
     * @return string Tag name
     */
    function getCode()
    {
        return $this->_code;
    }
    // }}}

    // {{{ getData()
    /**
     * Return data of the subfield
     *
     * @return string data
     */
    function getData()
    {
        return $this->_data;
    }
    // }}}

    // {{{ __toString()
    /**
     * Return string representation of subfield
     *
     * @return string String representation
     */
    public function __toString()
    {
        $pretty = '[' . $this->getCode() . ']: ' . $this->getData();
        return $pretty;
    }
    // }}}

    // {{{ toRaw()
    /**
     * Return the USMARC representation of the subfield
     *
     * @return string USMARC representation
     */
    function toRaw()
    {
        $result = MARC_SUBFIELD_INDICATOR.$this->_code.$this->_data;
        return $result;
    }
    // }}}

    // {{{ setCode()
    /**
     * Sets code of the subfield
     *
     * @param string $code new code for the subfield
     * @return string code 
     */
    function setCode($code)
    {
        if ($code) {
            // could check more stringently; m/[a-Z]/ or the likes
            $this->_code = $code;
        }
        else {
            // code must be _something_; raise error
            return false;
        }
        return true;
    }
    // }}}

    // {{{ setData()
    /**
     * Sets data of the subfield
     *
     * @param string $data new data for the subfield
     * @return string data
     */
    function setData($data)
    {
        $this->_data = $data;
        return true;
    }
    // }}}

    // {{{ isEmpty()
    /**
     * Checks whether the subfield is empty or not
     *
     * @return bool True or false
     */
    function isEmpty()
    {
        // There is data
        if ($this->_data) {
            return false;
        }

        // There is no data
        return true;
    }
    // }}}
}
// }}}

// {{{ class MARC_Linked_List
/**
 * The MARC_Linked_List class represents a linked list structure composed of {@link MARC_Link} objects.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class MARC_Linked_List implements Iterator {
    // {{{ properties
    /**
     * Current node in the linked list
     * @var MARC_Link
     */
    protected $_current;

    /**
     * Root node of the linked list
     * @var MARC_Link
     */
    protected $_root_node;

    /**
     * Error stack
     *
     * @var PEAR::ErrorStack
     */
    protected $_stack;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * MARC_Linked_List constructor
     *
     * @param MARC_Link $marc root node for the linked list
     */
    function __construct($marc=null) {
        $this->_stack = new PEAR_ErrorStack('MARC_Linked_List');
        $this->_root_node = $marc;
        $this->_current = $this->_root_node;
    }
    // }}}

    // {{{ function current()
    /**
     * Returns the current node in the linked list
     *
     * @return MARC_Link current node in the linked list
     */
    public function current() {
        return $this->_current;
    }
    // }}}

    // {{{ function rewind()
    /**
     * Sets the pointer for the linked list to the root node
     *
     * @return MARC_Link root node in the linked list
     */
    public function rewind() {
        $this->_current = $this->_root_node;
        return $this->_current;
    }
    // }}}

    // {{{ function key()
    /**
     * Stub for Iterator interface that simply returns the current node
     *
     * @return MARC_Link current node in the linked list
     */
    public function key() {
        return $this->_current;
    }
    // }}}

    // {{{ function key()
    /**
     * Stub for Iterator interface that simply returns the current node
     *
     * @return MARC_Link current node in the linked list
     */
    public function valid() {
        return $this->current();
    }
    // }}}

    // {{{ function next()
    /**
     * Sets the pointer for the linked list to the next link and
     * returns that link
     *
     * @return MARC_Link next node in the linked list
     */
    public function next() {
        if (!$this->current()) {
            return false;
        }
        $this->_current =& $this->current()->next();
        return $this->current();
    }
    // }}}

    // {{{ function previous()
    /**
     * Sets the pointer for the linked list to the previous link and
     * returns that link
     *
     * @return MARC_Link previous node in the linked list
     */
    public function previous() {
        if (!$this->current()->previous()) {
            return false;
        }
        $this->_current =& $this->current()->previous();
        return $this->current();
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $this->_stack;
    }
    // }}}

    // {{{ function addLink()
    /**
     * Adds a @see MARC_Link object to the linked list.
     *
     * While the @see MARC_Link constructor accepts a reference link
     * as the previous link in the chain, you might want to create
     * an extension of the @see MARC_Link object that enables you to
     * first create and manipulate the object before adding it to the
     * linked list.
     *
     * @param MARC_Link $newLink New link to add to the list
     * @param int $mode Specifies how the link will be inserted
     * @param MARC_Link $existingLink Reference link: if null, then use the final link in the list
     * @return bool Success or failure
     **/
    public function addLink($newLink, $mode=MARC_LINK_ADD_APPEND, &$existingLink=null) {
        // Try to get a valid value for the existing link
        if (!$existingLink) {
            /* We do not have a reference node for this link;
               try to use the current node in this linked list */
            $existingLink =& $this->rewind();
            
            if ($existingLink) {
                /* Iterate through to the end of the list */
                while ($x =& $this->next()) {
                    $existingLink = $x;
                }
            }
            else {
                /* This becomes the first link in the list */
                $this->__construct($newLink);
                return true;
            }
        }

        // Now add the link according to the requested mode
        switch ($mode) {

        case MARC_LINK_ADD_BEFORE:
            $previousLink =& $existingLink->previous();
            if ($previousLink) {
                $previousLink->setNext($newLink);
                $newLink->setPrevious($previousLink);
            }
            else {
                /* Existing link must be root node; make new link root */
                $this->_root_node = $newLink;
                $newLink->setPrevious();
            }
            /* We do these for both cases */
            $newLink->setNext($existingLink);
            $existingLink->setPrevious($newLink);

            break;

        case MARC_LINK_ADD_APPEND:
            if ($existingLink) {
                /* Iterate through to the end of the list */
                while ($x =& $this->next()) {
                    $existingLink = $x;
                }
            }

            /* Fall through deliberately to MARC_LINK_ADD_AFTER */

        case MARC_LINK_ADD_AFTER:
            $newLink->setPrevious($existingLink);
            $newLink->setNext($existingLink->next());
            $existingLink->setNext($newLink);
            break;

        default:
            return false;
            break;

        }

        return true;
    }
    // }}}

    // {{{ deleteLink()
    /**
     * @param {@link MARC_Subfield} $subfield
     */
    function deleteLink($subfield)
    {
        if ($subfield == $this->_root_node) {
            $this->_root_node = $subfield->next();
        }
        $subfield->delete();
    }
    // }}}

}
// }}}

// {{{ class MARC_Link
/**
 * The MARC_Link class represents a node in a {@link MARC_Linked_List} linked list structure.
 *
 * @category   File Formats
 * @package    File_MARC
 * @author     Dan Scott <dscott@laurentian.ca>
 * @license    http://www.gnu.org/copyleft/lesser.html  LGPL License 2.1
 * @link       http://pear.php.net/package/File_MARC
 */
class MARC_Link {
    // {{{ properties
    /**
     * Next node in the linked list
     * @var MARC_Link
     */
    protected $_next;

    /**
     * Previous node in the linked list
     * @var MARC_Link
     */
    protected $_previous;

    /**
     * Error stack
     *
     * @var PEAR::ErrorStack
     */
    protected $_stack;

    // }}}

    // {{{ Constructor: function __construct()
    /**
     * MARC_Link constructor
     * 
     * @param MARC_Link $link existing {@link MARC_Link} to use as the previous link in the list
     */
    function __construct($link=null) {
        $this->_stack = new PEAR_ErrorStack('MARC_Link');
        if ($link) {
            /* Make the passed-in link the previous link for this node */
            $this->setPrevious($link);
            /* Make the next link for the passed-in link the next link for this node instead */
            $this->setNext($link->next());
            /* Make this node the passed-in link's next link */
            $link->setNext($this);
        }
        else {
            $this->_next = null;
            $this->_previous = null;
        }
    }
    // }}}

    // {{{ Destructor: function __destruct()
    /**
     * Removes link from the list, adjusting the next and previous links accordingly
     *
     * This is a problem if the link is the root node for the list.
     * At this point, however, we do not have access to the list itself. Hmm.
     */
    function __destruct() {
        $next =& $this->next();
        $previous =& $this->previous();
        if ($previous) {
            $previous->setNext($next);
        }
        if ($next) {
            $next->setPrevious($previous);
        }
        $this->_next = null;
        $this->_previous = null;
        $this->_stack = null;
    }
    // }}}

    // {{{ function next()
    /**
     * Return the next link in the linked list
     *
     * @return MARC_Link next node in the linked list
     */
    public function next() {
        return $this->_next;
    }
    // }}}

    // {{{ function previous()
    /**
     * Return the previous link in the linked list
     *
     * @return MARC_Link previous node in the linked list
     */
    public function previous() {
        return $this->_previous;
    }
    // }}}

    // {{{ getErrors()
    /**
     * Get errors (if any) for this object 
     *
     * Returns the error stack.
     *
     * @return PEAR_ErrorStack returns the error stack
     */
    function getErrors()
    {
        return $this->_stack;
    }
    // }}}

    // {{{ function setNext()
    /**
     * Sets the pointer for the next link in the linked list to the specified link
     *
     * @param MARC_Link $link new next node in the linked list
     * @return MARC_Link new next node in the linked list
     */
    public function setNext(&$link=null) {
        $this->_next = $link;
        return $this->_next;
    }
    // }}}

    // {{{ function setPrevious()
    /**
     * Sets the pointer for the previous link in the linked list to the specified link
     *
     * @param MARC_Link $link new previous node in the linked list
     * @return MARC_Link new previous node in the linked list
     */
    public function setPrevious(&$link=null) {
        $this->_previous = $link;
        return $this->_previous;
    }
    // }}}
}
// }}}

?>

